import lecture1.parallel

def numContributeTo(input: Array[Int], kernel: Array[Int], output: Array[Int], i: Int): Int =
  if i < 0 || i >= output.length then
    0
  else if (i - kernel.length < 0) then
    i+1
  else if (i >= input.length) then
    output.length - i
  else kernel.length

def sequentialConvolve(
  input: Array[Int],
  kernel: Array[Int],
  output: Array[Int],
  from: Int,
  until: Int
  ): Unit = {
    var iOutput = from
    while iOutput < until do
      var iKernel = math.max(0, iOutput - input.length + 1)
      while iKernel < kernel.length && iKernel <= iOutput do
        output(iOutput) += input(iOutput - iKernel) * kernel(iKernel)
        iKernel += 1
      iOutput += 1
  }

def parallelConvolve(
    input: Array[Int],
    kernel: Array[Int],
    output: Array[Int],
    from: Int,
    until: Int,
    threshold: Int
): Unit = 
  if (until - from) <= threshold then
    sequentialConvolve(input, kernel, output, from, until)
  else
    val mid = from + (until - from) / 2
    parallel(
      parallelConvolve(input, kernel, output, from, mid, threshold),
      parallelConvolve(input, kernel, output, mid, until, threshold)
    )

object Original_WithOutputRace:

  def sequentialConvolve(
      input: Array[Int],
      kernel: Array[Int],
      output: Array[Int],
      from: Int,
      until: Int): Unit =
    var iOutput = from
    while iOutput < until do
      var iKernel = 0
      while iKernel < kernel.length do
        val iInput = iOutput - iKernel
        if 0 <= iInput && iInput < input.length then
          output(iOutput) += input(iInput) * kernel(iKernel)
        iKernel += 1
      iOutput += 1

  def parallelConvolve(
      input: Array[Int],
      kernel: Array[Int],
      output: Array[Int],
      from: Int,
      until: Int,
      threshold: Int): Unit =
    if (until - from) <= threshold then
      sequentialConvolve(input, kernel, output, from, until)
    else
      val mid = from + (until - from)/2
      parallel(
        parallelConvolve(input, kernel, output, from, mid, threshold),
        parallelConvolve(input, kernel, output, mid, until, threshold))
